/***********************************************************************

  AVMODE.C

  Notes:
    Command-line syntax is:
      AVMODE <character columns> <char width> <char height> [adjust]

    WARNING: This program may not function correctly in some SVGA
    subsystems.  Reasons why the program may fail:

      - SVGA write-protection of VGA CRTC registers
      - vertical or horizontal timings that differ from those used in
	IBM VGA BIOS video modes

***********************************************************************/

#include    <conio.h>	    /* _inp _outp */
#include    <dos.h>	    /* _int86 */
#include    <math.h>	    /* fabs */
#include    <stdio.h>	    /* printf etc. */
#include    <vs2.h>

#define DOTRATE0	25.175E6     /* default in VGA 350-line modes */
#define DOTRATE1	28.322E6     /* default in VGA 400-line modes */


/***********************************************************************

  FUNCTION PROTOTYPES

***********************************************************************/

int	    main( int, char * * );
static void SetCRTCReg( int, int );
static int  GetCRTCReg( int );
static void SetSeqReg( int, int );
static void SetATCReg( int, int );


/***********************************************************************

  GLOBAL VARIABLES

***********************************************************************/

static union _REGS  regs;
static LPVDDA	    lpVDDA = ADDRVDDA;


/***********************************************************************

  main

***********************************************************************/

int main( int argc, char * argv[] )
{
    double  DotRate;
    double  LineRate;
    double  FrameRate;
    double  HSyncDuration = 3.813E-6;	  /* duration of HSync signal */
    USHORT  CharWidth, CharHeight;

    USHORT  Cols;	     /* number of displayed character columns */
    int     HSyncTime;		       /* number of horiz sync clocks */
    USHORT  HPixels;		       /* number of horizontal pixels */
    int     HOverscan = 8;		 /* number of overscan clocks */
    int     HTotal;		    /* total clocks in one horiz scan */
    int     HAdjust = 0;		     /* horizontal adjustment */

    int     CRTCval[0x19];		      /* values for CRTC regs */
    int     i, j;
    UCHAR   SaveArea[16];


    /* verify the command line */
    if( argc < 4 )
    {
      printf( "\nSyntax: AVMODE" );
      printf( " <char cols> <char width> <char height> [adjust]" );
      return 3;
    }

    sscanf( argv[1], "%d", &Cols );
    sscanf( argv[2], "%d", &CharWidth );
    sscanf( argv[3], "%d", &CharHeight );
    if( argc == 5 )
      sscanf( argv[4], "%d", &HAdjust );

    /* use 400-line alphanumeric mode when setting the video mode */
    regs.h.ah = 0x12;
    regs.h.bl = 0x30;
    regs.h.al = 0x02;
    _int86( 0x10, &regs, &regs );

    /* start with BIOS 80x25 alphanumeric mode */
    regs.x.ax = 0x0003;
    _int86( 0x10, &regs, &regs );

    /* null the first alphanumeric character-generator glyph table */
    ExposeCGenRAM( &SaveArea );
    for( i=0; i<256; i++ )
      for( j=0; j<32; j++ )
	*(LPCHAR)(0xA0000000L + i*32 + j) = 0;
    HideCGenRAM( &SaveArea );

    /* call BIOS to select the vertical size of displayed characters */
    regs.h.ah = 0x11;
    regs.h.bl = 0;

    if( CharHeight < 13 )
      regs.h.al = 0x12; 	     /* use 8x8 character definitions */
    else
    if( CharHeight < 15 )
      regs.h.al = 0x11; 	    /* use 8x14 character definitions */
    else
      regs.h.al = 0x14; 	    /* use 8x16 character definitions */
    _int86( 0x10, &regs, &regs );

    /* set displayed character width (8 or 9 dots) */
    _outp( 0x3C4, 1 );	  /* update Sequencer Clocking Mode reg bit 0 */
    i = _inp( 0x3C5 );
    i = (i & 0xFE) | ( (CharWidth == 8) ? 1 : 0 );

    VRWait();
    SetSeqReg( 0, 1 );		       /* synchronous Sequencer reset */
    SetSeqReg( 1, i );			  /* update Clocking Mode reg */
    SetSeqReg( 0, 3 );			     /* clear Sequencer reset */

    j = ( (CharWidth == 8) ? 0 : 8 );  /* update Attribute Controller */
    SetATCReg( 0x13, j );		   /*  Horizontal Pel Pan reg */

    /* select the dot clock */
    i = _inp( 0x3CC );		     /* read Miscellaneous Output reg */
    i &= 0xF3;			       /* bits 2-3 = 00b (25.172 MHz) */

    /* select the dot clock based on which one gives a */
    /*	horizontal scan rate (line rate) closest to 31.5 KHz */
    HPixels = CharWidth * Cols;
    if( fabs(DOTRATE1/(double)(HPixels+160) - 31.5E3) <
	    fabs(DOTRATE0/(double)(HPixels+160) - 31.5E3) )
    {
      i |= 4;			       /* bits 2-3 = 01b (28.322 MHz) */
      DotRate = DOTRATE1;
    }

    else
      DotRate = DOTRATE0;

    VRWait();
    SetSeqReg( 0, 1 );		       /* synchronous Sequencer reset */
    _outp( 0x3C2, i );			    /* update Misc Output reg */
    SetSeqReg( 0, 3 );			     /* clear Sequencer reset */

    HSyncTime = 	   /* duration of Horiz Sync signal (rounded) */
      (int)(((DotRate * HSyncDuration) / (float)CharWidth) + 0.5);

    HTotal = Cols + HSyncTime + HOverscan;

    for( i=0; i<=0x18; i++ )	  /* get current CRTC register values */
      CRTCval[i] = GetCRTCReg( i );

    CRTCval[0] = HTotal - 5;			/* Reg 0: Horiz Total */
    CRTCval[1] = Cols - 1;	      /* Reg 1: Horiz Disp Enable End */
    CRTCval[2] = Cols;			  /* Reg 2: Start Horiz Blank */
    CRTCval[3] = ((HTotal-2) & 0x1F) | 0x80; /* Reg 3: End Horz Blank */
    CRTCval[4] = Cols + HAdjust + 4;		/* Reg 4: Start Hsync */
    CRTCval[5] =				  /* Reg 5: End Hsync */
      ((CRTCval[4] + HSyncTime) & 0x1F) |		  /* bits 0-4 */
      (((HTotal-2) & 0x20) << 2);			     /* bit 7 */
    CRTCval[9] = (CRTCval[9] & 0xE0) +	      /* Reg 9: Max Scan Line */
      CharHeight - 1;
    CRTCval[0x13] = Cols / 2;			  /* Reg 0x13: Offset */

    /* program the CRTC with new timing parameters */
    CRTCval[0x11] &= 0x7F;		/* disable CRTC write-protect */
    SetCRTCReg( 0x11, CRTCval[0x11] );

    for( i=0; i<=0x18; i++ )
      SetCRTCReg( i, CRTCval[i] );		  /* update CRTC regs */

    CRTCval[0x11] |= 0x80;		 /* enable CRTC write-protect */
    SetCRTCReg( 0x11, CRTCval[0x11] );

    SetATCReg( 0x11, 1 );			 /* set a blue border */

    /* update BIOS data area */
    lpVDDA->CRTCols = Cols;
    lpVDDA->Rows = (UCHAR)(400/CharHeight - 1);
    lpVDDA->CRTLen =
	(USHORT)((((lpVDDA->Rows + 1) * Cols * 2) + 0xFF) & 0xFF00);
    lpVDDA->Points = CharHeight;

    /* display register values and timings */
    printf( "\nDisplayed resolution is %d by %d characters",
      lpVDDA->CRTCols, lpVDDA->Rows+1 );
    printf( "\nDisplayed characters are %d by %d pixels",
      CharWidth, CharHeight );

    printf( "\n\nCRTC registers:\n" );	 /* show CRTC register values */
    for( i=0; i<=0x18; i++ )
      printf( " %2X", i );

    printf( "\n" );
    for( i=0; i<=0x18; i++ )
      printf( " %2X", GetCRTCReg( i ) );

    LineRate =				     /* scan lines per second */
      DotRate / (float)((CRTCval[0]+5)*CharWidth);
    i = ((GetCRTCReg(7) & 0x20) << 4)  /* vertical total (scan lines) */
	+ ((GetCRTCReg(7) & 0x1) << 8)
	+ GetCRTCReg(6) + 2;
    FrameRate = LineRate / (float)i;		 /* frames per second */

    printf( "\n\nDot rate   = %6.3f MHz, %d dots/char",
	DotRate/1E6, CharWidth );
    printf( "\nLine rate  = %5.2f KHz", LineRate/1E3 );
    printf( "\nFrame rate = %4.1f Hz\n", FrameRate );

    return 0;
}


/***********************************************************************

  SetCRTCReg

***********************************************************************/

static void SetCRTCReg( int r, int v )
{
    _outp( lpVDDA->Addr6845, r );
    _outp( lpVDDA->Addr6845+1, v );
}


/***********************************************************************

  GetCRTCReg

***********************************************************************/

static int GetCRTCReg( int r )
{
    _outp( lpVDDA->Addr6845, r );
    return _inp( lpVDDA->Addr6845+1 );
}


/***********************************************************************

  SetSeqReg

***********************************************************************/

static void SetSeqReg( int r, int v )
{
    _outp( 0x3C4, r );
    _outp( 0x3C5, v );
}


/***********************************************************************

  SetATCReg

***********************************************************************/

static void SetATCReg( int r, int v )
{
    regs.x.ax = 0x1000;
    regs.x.bx = (v << 8) + r;
    _int86( 0x10, &regs, &regs );
}
