/***********************************************************************

  GVMODE.C

  Notes:
    Command-line syntax is:
      GVMODE <horizontal pixels> <vertical pixels> [adjust]

    WARNING: This program may not function correctly in some SVGA
    subsystems.  Reasons why the program may fail:

      - SVGA write-protection of VGA CRTC registers
      - vertical or horizontal timings that differ from those used in
	IBM VGA BIOS video modes

***********************************************************************/

#include    <conio.h>	    /* _inp _outp */
#include    <dos.h>	    /* _int86 */
#include    <math.h>	    /* fabs */
#include    <stdio.h>	    /* printf etc. */
#include    <stdlib.h>	    /* __min */
#include    <vs2.h>

#define	DOTRATE0	25.175E6
#define	DOTRATE1	28.322E6
#define CHARHEIGHT	16		  /* (should be 8, 14, or 16) */


/***********************************************************************

  FUNCTION PROTOTYPES

***********************************************************************/

int	    main( int, char * * );
static void SetCRTCReg( int, int );
static int  GetCRTCReg( int );
static void SetSeqReg( int, int );
static void SetATCReg( int, int );


/***********************************************************************

  GLOBAL VARIABLES

***********************************************************************/

static union _REGS  regs;
static LPVDDA	    lpVDDA = ADDRVDDA;


/***********************************************************************

  main

  Notes:
    Typical values for HSync and VSync signals are based on IBM VGA.

***********************************************************************/

int main( int argc, char * argv[] )
{
    double  DotRate;
    double  LineRate;
    double  FrameRate;
    double  HSyncDuration = 3.813E-6;	  /* duration of HSync signal */
    double  VSyncDuration = 0.064E-3;	  /* duration of VSync signal */
    USHORT  CharWidth = 8;	      /* size of displayed characters */
    int     CharHeight = CHARHEIGHT;

    USHORT  Cols;		       /* displayed character columns */
    int     HSyncTime;		       /* number of horiz sync clocks */
    USHORT  HPixels;		       /* number of horizontal pixels */
    int     HBlank = 20;	 /* number of horizontal-blank clocks */
				     /* (includes overscan and hsync) */
    int     HTotal;		    /* total clocks in one horiz scan */
    int     HAdjust = 0;		     /* horizontal adjustment */

    int     VSyncTime;	       /* # of scan lines of vert sync signal */
    USHORT  VPixels;	    /* number of vertical pixels (scan lines) */
    int     VBlank = 45;       /* number of vertical-blank scan lines */
				     /* (includes overscan and vsync) */
    int     VTotal;		     /* total scan lines in one frame */

    int     CRTCval[0x19];		      /* values for CRTC regs */
    int     i;
	

    /* verify the command line */
    if( argc < 3 )
    {
      printf( "\nSyntax: GVMODE" );
      printf( " <horiz pixels> <vert pixels> [adjust]" );
      return 3;
    }

    sscanf( argv[1], "%d", &HPixels );
    if( HPixels % 16 )
    {
      printf( "\nError: Horizontal pixel count must be"
	      " a multiple of 16" );
      return 3;
    }

    sscanf( argv[2], "%d", &VPixels );
    if( argc == 4 )
      sscanf( argv[3], "%d", &HAdjust );

    /* start with either 640x350 or 640x480 16-color graphics mode */
    regs.x.ax = ( (VPixels <= 350) ? 0x0010 : 0x0012 );
    _int86( 0x10, &regs, &regs );

    /* call the BIOS to select the displayed character matrix */
    regs.h.ah = 0x11;
    switch( CharHeight )
    {
      case 8:
	regs.h.al = 0x23;	     /* use 8x8 character definitions */
	break;

      case 14:
	regs.h.al = 0x22;	    /* use 8x14 character definitions */
	break;

      case 16:
	regs.h.al = 0x24;	    /* use 8x16 character definitions */
	break;
    }

    regs.x.bx = 0;
    regs.x.dx = VPixels / CharHeight;	  /* number of character rows */
    _int86( 0x10, &regs, &regs );

    /* select the dot clock */
    i = _inp( 0x3CC );		     /* read Miscellaneous Output reg */
    i &= 0xF3;			       /* bits 2-3 = 00b (25.172 MHz) */

    /* select the dot clock based on which one gives a */
    /*	horizontal scan rate (line rate) closest to 31.5 KHz */
    if( fabs(DOTRATE1/(double)(HPixels+160) - 31.5E3) <
	    fabs(DOTRATE0/(double)(HPixels+160) - 31.5E3) )
    {
      i |= 4;			       /* bits 2-3 = 01b (28.322 MHz) */
    }

    VRWait();
    SetSeqReg( 0, 1 );		       /* synchronous Sequencer reset */
    _outp( 0x3C2, i );			    /* update Misc Output reg */
    SetSeqReg( 0, 3 );			     /* clear Sequencer reset */

    /* compute CRTC horizontal and vertical timing parameters */
    Cols = (USHORT)(HPixels / CharWidth);
    HTotal = Cols + HBlank;
    VTotal = VPixels + VBlank;

    DotRate = ( ((_inp(0x3CC) & 0x0C) == 0) ? DOTRATE0 : DOTRATE1);
    LineRate =				     /* scan lines per second */
      DotRate / (double)(HTotal*CharWidth);

    HSyncTime = 		     /* duration of Horiz Sync signal */
      (int)( (DotRate * HSyncDuration) / (double)CharWidth ) + 1;
    VSyncTime = (int)(VSyncDuration * LineRate);

    for( i=0; i<=0x18; i++ )	  /* get current CRTC register values */
      CRTCval[i] = GetCRTCReg( i );

    CRTCval[0] = HTotal - 5;			/* Reg 0: Horiz Total */
    CRTCval[1] = Cols - 1;	      /* Reg 1: Horiz Disp Enable End */
    CRTCval[2] = Cols;			  /* Reg 2: Start Horiz Blank */
    CRTCval[3] = ((HTotal-2) & 0x1F);	    /* Reg 3: End Horiz Blank */
    CRTCval[4] = Cols + ((HBlank-HSyncTime)/2) + HAdjust;
						/* Reg 4: Start Hsync */
    CRTCval[5] =				  /* Reg 5: End Hsync */
      ((CRTCval[4] + HSyncTime) & 0x1F) |		  /* bits 0-4 */
      (((HTotal-2) & 0x20) << 2);			     /* bit 7 */
    CRTCval[0x13] = Cols / 2;			  /* Reg 0x13: Offset */

    CRTCval[6] =			     /* Reg 6: Vertical Total */
      VPixels + VBlank - 2;
    CRTCval[0x10] =		      /* Reg 0x10: Vert Retrace Start */
      VPixels + (VBlank-VSyncTime)/2;
    CRTCval[0x11] =			/* Reg 0x11: Vert Retrace End */
      (CRTCval[0x11] & 0xF0) |				  /* bits 4-7 */
      ((CRTCval[0x10] + VSyncTime) & 0x0F);		  /* bits 0-3 */
    CRTCval[0x12] = VPixels - 1; /* Reg 0x12: Vert Display Enable End */
    CRTCval[0x15] =			/* Reg 0x15: Start Vert Blank */
      VPixels + __min(8,VBlank/2) - 1;
    CRTCval[0x16] =			  /* Reg 0x16: End Vert Blank */
      (VPixels + VBlank) - __min(8,VBlank/2);

    CRTCval[7] =				   /* Reg 7: Overflow */
      ((CRTCval[0x10] & 0x200) >> 2) |			     /* bit 7 */
      ((CRTCval[0x12] & 0x200) >> 3) |			     /* bit 6 */
      ((CRTCval[6] & 0x200) >> 4) |			     /* bit 5 */
      (CRTCval[7] & 0x10) |				     /* bit 4 */
      ((CRTCval[0x15] & 0x100) >> 5) |			     /* bit 3 */
      ((CRTCval[0x10] & 0x100) >> 6) |			     /* bit 2 */
      ((CRTCval[0x12] & 0x100) >> 7) |			     /* bit 1 */
      ((CRTCval[6] & 0x100) >> 8);			     /* bit 0 */
    CRTCval[9] =			      /* Reg 9: Max Scan Line */
      (CRTCval[9] & 0xDF) | ((CRTCval[0x15] & 0x200) >> 4);  /* bit 5 */

    /* program the CRTC with new timing parameters */
    CRTCval[0x11] &= 0x7F;		/* disable CRTC write-protect */
    SetCRTCReg( 0x11, CRTCval[0x11] );

    for( i=0; i<=0x18; i++ )
      SetCRTCReg( i, CRTCval[i] );		  /* update CRTC regs */

    CRTCval[0x11] |= 0x80;		 /* enable CRTC write-protect */
    SetCRTCReg( 0x11, CRTCval[0x11] );

    SetATCReg( 0x11, 1 );			 /* set a blue border */

    /* update BIOS data area */
    lpVDDA->CRTCols = Cols;
    lpVDDA->CRTLen = (USHORT)(((VPixels * Cols * 2) + 0xFFF) & 0xF000);

    /* show register values and timings */
    printf( "\nDisplayed resolution is %d by %d pixels",
	HPixels, VPixels );
    printf( "\nDisplayed character matrix is %d by %d pixels",
	CharWidth, CharHeight );

    printf( "\n\nCRTC registers:\n" );	 /* show CRTC register values */
    for( i=0; i<=0x18; i++ )
      printf( " %2X", i );

    printf( "\n" );
    for( i=0; i<=0x18; i++ )
      printf( " %2X", GetCRTCReg( i ) );

    i = ((GetCRTCReg(7) & 0x20) << 4)  /* vertical total (scan lines) */
	+ ((GetCRTCReg(7) & 0x1) << 8)
	+ GetCRTCReg(6) + 2;
    FrameRate = LineRate / (double)i;		 /* frames per second */

    printf( "\n\nDot rate = %6.3f MHz, %d dots/char",
	DotRate/1E6, CharWidth );
    printf( "\nLine rate  = %6.3f KHz", LineRate/1E3 );
    printf( "\nFrame rate = %5.2f Hz\n", FrameRate );
    printf( "\nVertical blanking interval = %5.2f ms\n",
		 (1000.0 * VBlank / FrameRate) / VTotal );
    return 0;
}


/***********************************************************************

  SetCRTCReg

***********************************************************************/

static void SetCRTCReg( int r, int v )
{
    _outp( lpVDDA->Addr6845, r );
    _outp( lpVDDA->Addr6845+1, v );
}


/***********************************************************************

  GetCRTCReg

***********************************************************************/

static int GetCRTCReg( int r )
{
    _outp( lpVDDA->Addr6845, r );
    return _inp( lpVDDA->Addr6845+1 );
}


/***********************************************************************

  SetSeqReg

***********************************************************************/

static void SetSeqReg( int r, int v )
{
    _outp( 0x3C4, r );
    _outp( 0x3C5, v );
}


/***********************************************************************

  SetATCReg

***********************************************************************/

static void SetATCReg( int r, int v )
{
    regs.x.ax = 0x1000;
    regs.x.bx = (v << 8) + r;
    _int86( 0x10, &regs, &regs );
}
